"""Delete unused files from jobbergate's file storage."""

import asyncio

from fastapi import BackgroundTasks
from loguru import logger
from sqlalchemy import select


async def get_set_of_files_from_database(session, table) -> set[str]:
    """Get a set of files from the database."""
    rows = await session.execute(select(table))
    result = {obj.file_key for obj in rows.scalars().all()}
    logger.debug(f"Total of files found in the table {table.__tablename__}: {len(result)}")
    return result


async def get_set_of_files_from_bucket(bucket, table) -> set[str]:
    """Get a set of files from the bucket."""
    prefix = table.__tablename__
    result = {obj.key async for obj in bucket.objects.filter(Prefix=prefix)}
    logger.debug(f"Total of files found in the bucket {bucket.name} with prefix {prefix}: {len(result)}")
    return result


async def get_files_to_delete(session, table, bucket) -> set[str]:
    """Get a set of files to delete."""
    files_in_database = await get_set_of_files_from_database(session, table)
    files_in_bucket = await get_set_of_files_from_bucket(bucket, table)
    result = files_in_bucket - files_in_database
    logger.debug(f"Total of files to be garbage collected: {len(result)}")
    return result


async def delete_files_from_bucket(bucket, files_to_delete: set[str]) -> None:
    """Delete files from the bucket."""
    MAX_CONCURRENT_REQUESTS = 25
    semaphore = asyncio.Semaphore(MAX_CONCURRENT_REQUESTS)

    async def delete_file(file: str) -> None:
        async with semaphore:
            obj = await bucket.Object(file)
            await obj.delete()
        logger.debug(f"Deleted file {file} from bucket {bucket.name}")

    tasks = [delete_file(file) for file in files_to_delete]
    await asyncio.gather(*tasks)


async def garbage_collect(session, bucket, list_of_tables, background_tasks: BackgroundTasks) -> None:
    """Delete unused files from jobbergate's file storage."""
    for table in list_of_tables:
        files_to_delete = await get_files_to_delete(session, table, bucket)
        if files_to_delete:
            background_tasks.add_task(delete_files_from_bucket, bucket, files_to_delete)



"""Router for the Job Script Template resource."""

from typing import cast

from buzz import handle_errors
from fastapi import APIRouter, BackgroundTasks, Depends, File, HTTPException, Path, Query
from fastapi import Response as FastAPIResponse
from fastapi import UploadFile, status
from fastapi_pagination import Page
from loguru import logger

from jobbergate_api.apps.constants import FileType
from jobbergate_api.apps.dependencies import SecureService, secure_services
from jobbergate_api.apps.garbage_collector import garbage_collect
from jobbergate_api.apps.job_script_templates.models import JobScriptTemplate
from jobbergate_api.apps.job_script_templates.tools import coerce_id_or_identifier
from jobbergate_api.apps.job_scripts.models import JobScriptFile
from jobbergate_api.apps.job_scripts.schemas import (
    JobScriptBaseView,
    JobScriptCloneRequest,
    JobScriptCreateRequest,
    JobScriptDetailedView,
    JobScriptFileDetailedView,
    JobScriptListView,
    JobScriptUpdateRequest,
    RenderFromTemplateRequest,
)
from jobbergate_api.apps.job_scripts.tools import inject_sbatch_params
from jobbergate_api.apps.permissions import Permissions, can_bypass_ownership_check
from jobbergate_api.apps.schemas import ListParams
from jobbergate_api.apps.services import ServiceError

router = APIRouter(prefix="/job-scripts", tags=["Job Scripts"])


@router.delete(
    "/clean-unused-entries",
    status_code=status.HTTP_202_ACCEPTED,
    description="Endpoint to automatically clean unused job scripts depending on a threshold",
    tags=["Garbage collector"],
)
def job_script_auto_clean_unused_entries(
    background_tasks: BackgroundTasks,
    secure_services: SecureService = Depends(
        secure_services(Permissions.ADMIN, Permissions.JOB_SCRIPTS_DELETE)
    ),
):
    """Automatically clean unused job scripts depending on a threshold."""
    logger.info("Starting automatically cleanup for unused job scripts")
    background_tasks.add_task(secure_services.crud.job_script.auto_clean_unused_job_scripts)
    return {"description": "Automatically cleanup started"}


@router.post(
    "",
    status_code=status.HTTP_201_CREATED,
    response_model=JobScriptBaseView,
    description="Endpoint for creating a stand alone job script. Use file upload to add files.",
)
async def job_script_create(
    create_request: JobScriptCreateRequest,
    secure_services: SecureService = Depends(
        secure_services(Permissions.ADMIN, Permissions.JOB_SCRIPTS_CREATE)
    ),
):
    """Create a stand alone job script."""
    logger.info(f"Creating a new job script with {create_request=}")

    if secure_services.identity_payload.email is None:
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail="The token payload does not contain an email",
        )

    return await secure_services.crud.job_script.create(
        owner_email=secure_services.identity_payload.email,
        **create_request.model_dump(exclude_unset=True),
    )


@router.post(
    "/clone/{id}",
    status_code=status.HTTP_201_CREATED,
    response_model=JobScriptDetailedView,
    description="Endpoint for cloning a job script to a new entry owned by the user",
)
async def job_script_clone(
    id: int = Path(),
    clone_request: JobScriptCloneRequest | None = None,
    secure_services: SecureService = Depends(
        secure_services(Permissions.ADMIN, Permissions.JOB_SCRIPTS_CREATE, ensure_email=True)
    ),
):
    """Clone a job script by its id."""
    logger.info(f"Cloning a job script from {id=} with {clone_request=}")

    if clone_request is None:
        clone_request = JobScriptCloneRequest()

    original_instance = await secure_services.crud.job_script.get(id, include_files=True)
    cloned_instance = await secure_services.crud.job_script.clone_instance(
        original_instance,
        owner_email=secure_services.identity_payload.email,
        **clone_request.model_dump(exclude_unset=True, exclude_none=True),
    )

    for file in original_instance.files:
        await secure_services.file.job_script.clone_instance(file, cloned_instance.id)

    # Get again to update the file data
    return await secure_services.crud.job_script.get(cloned_instance.id, include_files=True)


@router.post(
    "/render-from-template/{id_or_identifier}",
    status_code=status.HTTP_201_CREATED,
    response_model=JobScriptDetailedView,
    description="Endpoint for job script creation",
)
async def job_script_create_from_template(
    create_request: JobScriptCreateRequest,
    render_request: RenderFromTemplateRequest,
    id_or_identifier: str = Path(...),
    secure_services: SecureService = Depends(
        secure_services(Permissions.ADMIN, Permissions.JOB_SCRIPTS_CREATE, ensure_email=True)
    ),
):
    """Create a new job script from a job script template."""
    typed_id_or_identifier = coerce_id_or_identifier(id_or_identifier)
    logger.info(f"Creating a new job script from {typed_id_or_identifier=} with {create_request=}")

    base_template = cast(
        JobScriptTemplate, await secure_services.crud.template.get(typed_id_or_identifier, include_files=True)
    )

    required_map = render_request.template_output_name_mapping

    required_keys = set(required_map.keys())
    provided_keys = set(f.filename for f in base_template.template_files)
    missing_keys = required_keys - provided_keys
    if len(missing_keys) > 0:
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail="The following required files are missing for job template {}: {}".format(
                typed_id_or_identifier, ", ".join(missing_keys)
            ),
        )

    mapped_template_files = {
        new_filename: file
        for file in base_template.template_files
        if (new_filename := required_map.get(file.filename, None))
    }

    entrypoint_files = [f for f in mapped_template_files.values() if f.file_type == FileType.ENTRYPOINT]
    if len(entrypoint_files) != 1:
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail=f"Exactly one entrypoint file must be specified, got {len(entrypoint_files)}",
        )

    job_script = await secure_services.crud.job_script.create(
        owner_email=secure_services.identity_payload.email,
        parent_template_id=base_template.id,
        **create_request.model_dump(exclude_unset=True),
    )

    for new_filename, template_file in mapped_template_files.items():
        file_content = await secure_services.file.template.render(
            template_file,
            render_request.param_dict,
        )

        if template_file.file_type == FileType.ENTRYPOINT and render_request.sbatch_params:
            with handle_errors(
                "Failed to inject sbatch params into the entrypoint file",
                raise_exc_class=ServiceError,
                raise_kwargs=dict(status_code=status.HTTP_422_UNPROCESSABLE_ENTITY),
            ):
                file_content = inject_sbatch_params(file_content, render_request.sbatch_params)

        await secure_services.file.job_script.upsert(
            parent_id=job_script.id,
            filename=new_filename,
            upload_content=file_content,
            file_type=template_file.file_type,
        )

    return await secure_services.crud.job_script.get(job_script.id, include_files=True)


@router.get(
    "/{id}",
    description="Endpoint to return a job script by its id",
    response_model=JobScriptDetailedView,
)
async def job_script_get(
    id: int = Path(),
    secure_services: SecureService = Depends(
        secure_services(Permissions.ADMIN, Permissions.JOB_SCRIPTS_READ, commit=False)
    ),
):
    """Get a job script by id."""
    logger.info(f"Getting job script {id=}")
    return await secure_services.crud.job_script.get(id, include_files=True)


@router.get(
    "",
    description="Endpoint to return a list of job scripts",
    response_model=Page[JobScriptListView],
)
async def job_script_get_list(
    list_params: ListParams = Depends(),
    from_job_script_template_id: int | None = Query(
        None,
        description="Filter job-scripts by the job-script-template-id they were created from.",
    ),
    secure_services: SecureService = Depends(
        secure_services(Permissions.ADMIN, Permissions.JOB_SCRIPTS_READ, commit=False)
    ),
):
    """Get a list of job scripts."""
    logger.debug("Preparing to list job scripts")

    list_kwargs = list_params.model_dump(exclude_unset=True, exclude={"user_only"})
    list_kwargs["include_parent"] = True

    if from_job_script_template_id is not None:
        list_kwargs["parent_template_id"] = from_job_script_template_id
    if list_params.user_only:
        list_kwargs["owner_email"] = secure_services.identity_payload.email

    return await secure_services.crud.job_script.paginated_list(**list_kwargs)


@router.put(
    "/{id}",
    status_code=status.HTTP_200_OK,
    description="Endpoint to update a job script by id",
    response_model=JobScriptListView,
)
async def job_script_update(
    update_params: JobScriptUpdateRequest,
    id: int = Path(),
    secure_services: SecureService = Depends(
        secure_services(Permissions.ADMIN, Permissions.JOB_SCRIPTS_UPDATE, ensure_email=True)
    ),
):
    """Update a job script template by id or identifier."""
    logger.info(f"Updating job script {id=} with {update_params=}")
    instance = await secure_services.crud.job_script.get(id, include_parent=True)
    if not can_bypass_ownership_check(secure_services.identity_payload.permissions):
        secure_services.crud.job_script.ensure_attribute(
            instance, owner_email=secure_services.identity_payload.email
        )
    return await secure_services.crud.job_script.update(id, **update_params.model_dump(exclude_unset=True))


@router.delete(
    "/{id}",
    status_code=status.HTTP_204_NO_CONTENT,
    description="Endpoint to delete a job script by id",
)
async def job_script_delete(
    id: int = Path(...),
    secure_services: SecureService = Depends(
        secure_services(Permissions.ADMIN, Permissions.JOB_SCRIPTS_DELETE, ensure_email=True)
    ),
):
    """Delete a job script template by id or identifier."""
    logger.info(f"Deleting job script {id=}")
    instance = await secure_services.crud.job_script.get(id)
    if not can_bypass_ownership_check(secure_services.identity_payload.permissions):
        secure_services.crud.job_script.ensure_attribute(
            instance, owner_email=secure_services.identity_payload.email
        )
    await secure_services.crud.job_script.delete(id)
    return FastAPIResponse(status_code=status.HTTP_204_NO_CONTENT)


@router.get(
    "/{id}/upload/{file_name:path}",
    description="Endpoint to get a file from a job script",
)
async def job_script_get_file(
    id: int = Path(...),
    file_name: str = Path(...),
    secure_services: SecureService = Depends(
        secure_services(Permissions.ADMIN, Permissions.JOB_SCRIPTS_READ, commit=False)
    ),
):
    """
    Get a job script file.

    Note:
        See https://fastapi.tiangolo.com/advanced/custom-response/#streamingresponse
    """
    job_script_file = await secure_services.file.job_script.get(id, file_name)
    return FastAPIResponse(
        content=await secure_services.file.job_script.get_file_content(job_script_file),
        media_type="text/plain",
        headers={"filename": job_script_file.filename},
    )


@router.put(
    "/{id}/upload/{file_type}",
    status_code=status.HTTP_200_OK,
    description=(
        "Endpoint to upload a file to a job script. "
        "If a previous filename is provided, the file will be renamed from that. "
        "Upload file is optional in this scenario since the file content can be copied from previous file."
    ),
    response_model=JobScriptFileDetailedView,
)
async def job_script_upload_file(
    id: int = Path(...),
    file_type: FileType = Path(...),
    filename: str | None = Query(None, max_length=255),
    upload_file: UploadFile | None = File(None, description="File to upload"),
    previous_filename: str | None = Query(
        None, description="Previous name of the file in case a rename is needed", max_length=255
    ),
    secure_services: SecureService = Depends(
        secure_services(Permissions.ADMIN, Permissions.JOB_SCRIPTS_CREATE, ensure_email=True)
    ),
):
    """Upload a file to a job script."""
    # This is included for backwards compatibility with the previous implementation
    # where filename was recovered from the upload_file object
    filename = filename or getattr(upload_file, "filename")
    if not filename:
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail="Filename must be provided either as a query parameter or as part of the file upload",
        )

    logger.debug(f"Uploading file {filename=} to job script {id=} with {file_type=} and {previous_filename=}")

    job_script = await secure_services.crud.job_script.get(id)
    if not can_bypass_ownership_check(secure_services.identity_payload.permissions):
        secure_services.crud.job_script.ensure_attribute(
            job_script, owner_email=secure_services.identity_payload.email
        )

    return await secure_services.file.job_script.upsert(
        parent_id=job_script.id,
        filename=filename,
        upload_content=upload_file,
        previous_filename=previous_filename,
        file_type=file_type,
    )


@router.delete(
    "/{id}/upload/{file_name}",
    status_code=status.HTTP_200_OK,
    description="Endpoint to delete a file from a job script",
)
async def job_script_delete_file(
    id: int = Path(...),
    file_name: str = Path(...),
    secure_services: SecureService = Depends(
        secure_services(Permissions.ADMIN, Permissions.JOB_SCRIPTS_DELETE, ensure_email=True)
    ),
):
    """Delete a file from a job script template by id or identifier."""
    job_script = await secure_services.crud.job_script.get(id)
    if not can_bypass_ownership_check(secure_services.identity_payload.permissions):
        secure_services.crud.job_script.ensure_attribute(
            job_script, owner_email=secure_services.identity_payload.email
        )
    job_script_file = await secure_services.file.job_script.get(job_script.id, file_name)
    await secure_services.file.job_script.delete(job_script_file)


@router.delete(
    "/upload/garbage-collector",
    status_code=status.HTTP_202_ACCEPTED,
    description="Endpoint to delete all unused files from the job script file storage",
    tags=["Garbage collector"],
)
def job_script_garbage_collector(
    background_tasks: BackgroundTasks,
    secure_services: SecureService = Depends(
        secure_services(Permissions.ADMIN, Permissions.JOB_SCRIPTS_DELETE)
    ),
):
    """Delete all unused files from job scripts on the file storage."""
    logger.info("Starting garbage collection from jobbergate file storage")
    background_tasks.add_task(
        garbage_collect,
        secure_services.session,
        secure_services.bucket,
        [JobScriptFile],
        background_tasks,
    )
    return {"description": "Garbage collection started"}



"""Define app-wide, reusable pydantic schemas."""

import datetime
from typing import Annotated, Any, List, Type

import pendulum
from pydantic import BaseModel, ConfigDict, Field, GetCoreSchemaHandler
from pydantic_core import PydanticCustomError, core_schema

# Make both Pydantic and mypy happy:
LengthLimitedStr = Annotated[str, Field(max_length=255)]


class PydanticDateTime(pendulum.DateTime):
    """
    A `pendulum.DateTime` object. At runtime, this type decomposes into pendulum.DateTime automatically.

    This type exists because Pydantic throws a fit on unknown types.

    This code is borrowed and enhanced from the `pydantic-extra-types` module but provides conversion from
    standard datetimes as well.
    """

    __slots__: List[str] = []

    @classmethod
    def __get_pydantic_core_schema__(
        cls, source: Type[Any], handler: GetCoreSchemaHandler
    ) -> core_schema.CoreSchema:
        """
        Return a Pydantic CoreSchema with the Datetime validation.

        Args:
            source: The source type to be converted.
            handler: The handler to get the CoreSchema.

        Returns:
            A Pydantic CoreSchema with the Datetime validation.
        """
        return core_schema.no_info_wrap_validator_function(cls._validate, core_schema.datetime_schema())

    @classmethod
    def _validate(cls, value: Any, handler: core_schema.ValidatorFunctionWrapHandler) -> Any:
        """
        Validate the datetime object and return it.

        Args:
            value: The value to validate.
            handler: The handler to get the CoreSchema.

        Returns:
            The validated value or raises a PydanticCustomError.
        """
        if isinstance(value, pendulum.DateTime):
            return handler(value)

        if isinstance(value, datetime.datetime):
            return handler(pendulum.instance(value))

        try:
            return handler(pendulum.parse(value))
        except Exception as exc:
            raise PydanticCustomError("value_error", "value is not a valid timestamp") from exc


class TableResource(BaseModel):
    """
    Describes a base for table models that include basic, common info.
    """

    id: int
    name: str
    owner_email: str
    created_at: PydanticDateTime
    updated_at: PydanticDateTime
    is_archived: bool
    description: str | None = None

    model_config = ConfigDict(from_attributes=True)


class ListParams(BaseModel):
    """
    Describe the shared parameters for a list request.
    """

    sort_ascending: bool = True
    user_only: bool = False
    search: LengthLimitedStr | None = None
    sort_field: LengthLimitedStr | None = None
    include_archived: bool = False
    include_parent: bool = False



"""Provide a method for coercing id_or_identifier to a string or int."""


def coerce_id_or_identifier(id_or_identifier: str) -> int | str:
    """
    Determine whether the id_or_identifier should be a string or an integer.

    This is necessary because FastAPI no longer automatically converts path parameters to integers
    automatically if they may also be string values.
    """
    try:
        return int(id_or_identifier)
    except ValueError:
        return id_or_identifier



"""Database models for the cluster health resource."""

from pendulum.datetime import DateTime as PendulumDateTime
from sqlalchemy import DateTime, Integer, String
from sqlalchemy.orm import Mapped, mapped_column

from jobbergate_api.apps.models import Base, CommonMixin, TimestampMixin


class ClusterStatus(CommonMixin, TimestampMixin, Base):
    """Cluster status table definition."""

    client_id: Mapped[str] = mapped_column(String, primary_key=True)
    interval: Mapped[int] = mapped_column(Integer, nullable=False)
    last_reported: Mapped[PendulumDateTime] = mapped_column(
        DateTime(timezone=True),
        nullable=False,
        default=PendulumDateTime.utcnow,
    )

    @property
    def is_healthy(self) -> bool:
        """Return True if the last_reported time is before now plus the interval in seconds between pings."""
        return PendulumDateTime.utcnow().subtract(seconds=self.interval) <= self.last_reported



"""Provide a convenience class for managing job-script files."""

import re

from loguru import logger


def inject_sbatch_params(job_script_data_as_string: str, sbatch_params: list[str]) -> str:
    """
    Inject sbatch params into job script.

    Given the job script as job_script_data_as_string, inject the sbatch params in the correct location.
    """
    logger.debug("Preparing to inject sbatch params into job script")

    if not sbatch_params:
        logger.warning("Sbatch param list is empty")
        return job_script_data_as_string

    # Find the first non-blank, non-comment line
    match = re.search(r"^[^#\n]", job_script_data_as_string, re.MULTILINE)
    if match:
        insert_index = match.start()
    else:
        # If no such line is found, append at the end
        insert_index = len(job_script_data_as_string)

    inner_string = "# Sbatch params injected at rendering time\n"
    for parameter in sbatch_params:
        inner_string += f"#SBATCH {parameter}\n"
    else:
        inner_string += "\n"

    new_job_script_data_as_string = (
        job_script_data_as_string[:insert_index] + inner_string + job_script_data_as_string[insert_index:]
    )

    logger.debug("Done injecting sbatch params into job script")
    return new_job_script_data_as_string



"""
JobSubmission resource schema.
"""

from typing import Optional

from pydantic import BaseModel, ConfigDict, Field, NonNegativeInt, field_validator

from jobbergate_api.apps.job_scripts.schemas import JobScriptBaseView, JobScriptDetailedView
from jobbergate_api.apps.job_submissions.constants import JobSubmissionStatus, SlurmJobState
from jobbergate_api.apps.schemas import LengthLimitedStr, TableResource
from jobbergate_api.meta_mapper import MetaField, MetaMapper

job_submission_meta_mapper = MetaMapper(
    id=MetaField(
        description="The unique database identifier for the instance",
        example=101,
    ),
    created_at=MetaField(
        description="The timestamp for when the instance was created",
        example="2023-08-18T13:55:37.172285",
    ),
    updated_at=MetaField(
        description="The timestamp for when the instance was last updated",
        example="2023-08-18T13:55:37.172285",
    ),
    name=MetaField(
        description="The unique name of the job submission",
        example="test-job-submission-77",
    ),
    description=MetaField(
        description="A text field providing a human-friendly description of the job_submission",
        example="Submission for the Foo job on sample 99 using the bar variant",
    ),
    owner_email=MetaField(
        description="The email of the owner/creator of the instance",
        example="tucker@omnivector.solutions",
    ),
    job_script_id=MetaField(
        description="The foreign-key to the job_script from which this instance was created",
        example=71,
    ),
    execution_directory=MetaField(
        description="The directory on the cluster where the job should be executed",
        example="/home/someuser/job-data/test-77",
    ),
    slurm_job_id=MetaField(
        description="The id for the slurm job executing this job_submission",
        example="1883",
    ),
    client_id=MetaField(
        description="The client_id of the cluster where this job submission should execute",
        example="mega-cluster-1",
    ),
    status=MetaField(
        description=f"The status of the job submission. Must be one of {JobSubmissionStatus.pretty_list()}",
        example=JobSubmissionStatus.CREATED,
    ),
    report_message=MetaField(
        description="The report message received from cluster-agent when a job submission is rejected",
        example="Unrecognized SBATCH arguments",
    ),
    slurm_job_state=MetaField(
        description="The Slurm Job state as reported by the agent.example",
        example="PENDING",
    ),
    slurm_job_info=MetaField(
        description="Detailed information about the Slurm Job as reported by the agent",
        example="""
            JobId=2 JobName=apptainer-test
            UserId=ubuntu(1000) GroupId=ubuntu(1000) MCS_label=N/A
            Priority=4294901758 Nice=0 Account=(null) QOS=normal
            JobState=RUNNING Reason=None Dependency=(null)
            Requeue=1 Restarts=0 BatchFlag=1 Reboot=0 ExitCode=0:0
            RunTime=00:01:57 TimeLimit=UNLIMITED TimeMin=N/A
            SubmitTime=2024-01-29T17:42:15 EligibleTime=2024-01-29T17:42:15
            AccrueTime=2024-01-29T17:42:15
            StartTime=2024-01-29T17:42:15 EndTime=Unknown Deadline=N/A
            SuspendTime=None SecsPreSuspend=0 LastSchedEval=2024-01-29T17:42:15 Scheduler=Main
            Partition=compute AllocNode:Sid=10.122.188.182:1278
            ReqNodeList=(null) ExcNodeList=(null)
            NodeList=democluster
            BatchHost=democluster
            NumNodes=1 NumCPUs=6 NumTasks=6 CPUs/Task=1 ReqB:S:C:T=0:0:*:*
            ReqTRES=cpu=6,mem=3903M,node=1,billing=6
            AllocTRES=cpu=6,node=1,billing=6
            Socks/Node=* NtasksPerN:B:S:C=0:0:*:* CoreSpec=*
            MinCPUsNode=1 MinMemoryNode=0 MinTmpDiskNode=0
            Features=(null) DelayBoot=00:00:00
            OverSubscribe=OK Contiguous=0 Licenses=(null) Network=(null)
            Command=(null)
            WorkDir=/tmp
            StdErr=/home/ubuntu/democluster/job.%J.err
            StdIn=/dev/null
            StdOut=/home/ubuntu/democluster/job.%J.out
            Power=
        """,
    ),
    is_archived=MetaField(
        description="Indicates if the job submission has been archived.",
        example=False,
    ),
    sbatch_arguments=MetaField(
        description="The arguments used to submit the job to the slurm queue",
        example=["--exclusive", "--job-name=example-job"],
    ),
)


class JobSubmissionCreateRequest(BaseModel):
    """
    Request model for creating JobSubmission instances.
    """

    name: LengthLimitedStr
    description: Optional[LengthLimitedStr] = None
    job_script_id: NonNegativeInt
    slurm_job_id: Optional[NonNegativeInt] = None
    execution_directory: Optional[LengthLimitedStr] = None
    client_id: Optional[LengthLimitedStr] = None
    sbatch_arguments: Optional[list[LengthLimitedStr]] = Field(None, max_length=50)

    @field_validator("execution_directory", mode="before")
    @classmethod
    def empty_str_to_none(cls, v):
        """Ensure empty strings are converted to None to avoid problems with Path downstream."""
        return v or None

    model_config = ConfigDict(json_schema_extra=job_submission_meta_mapper)


class JobSubmissionUpdateRequest(BaseModel):
    """
    Request model for updating JobSubmission instances.
    """

    name: Optional[LengthLimitedStr] = None
    description: Optional[LengthLimitedStr] = None
    execution_directory: Optional[LengthLimitedStr] = None
    status: Optional[JobSubmissionStatus] = None

    @field_validator("execution_directory", mode="before")
    def empty_str_to_none(cls, v):
        """Ensure empty strings are converted to None to avoid problems with Path downstream."""
        return v or None

    model_config = ConfigDict(json_schema_extra=job_submission_meta_mapper)


class JobSubmissionBaseView(TableResource):
    """
    Base model to match the database for the JobSubmission resource.

    Omits parent relationship.
    """

    job_script_id: Optional[int] = None
    slurm_job_id: Optional[int] = None
    client_id: str
    status: JobSubmissionStatus
    slurm_job_state: Optional[SlurmJobState] = None

    model_config = ConfigDict(json_schema_extra=job_submission_meta_mapper)


class JobSubmissionListView(JobSubmissionBaseView):
    """
    Complete model to match the database for the JobSubmission resource in a list view.
    """

    job_script: Optional[JobScriptBaseView] = None


class JobSubmissionDetailedView(JobSubmissionBaseView):
    """
    Complete model to match the database for the JobSubmission resource in a detailed view.
    """

    execution_directory: Optional[str]
    report_message: Optional[str]
    slurm_job_info: Optional[str]
    sbatch_arguments: Optional[list[str]]


class PendingJobSubmission(BaseModel):
    """
    Specialized model for the cluster-agent to pull pending job_submissions.

    Model also includes data from its job_script and application sources.
    """

    id: int
    name: str
    owner_email: str
    execution_directory: Optional[str] = None
    execution_parameters: dict = Field(default_factory=dict)
    job_script: JobScriptDetailedView
    sbatch_arguments: Optional[list[str]] = None

    model_config = ConfigDict(
        from_attributes=True, extra="ignore", json_schema_extra=job_submission_meta_mapper
    )


class ActiveJobSubmission(BaseModel):
    """
    Specialized model for the cluster-agent to pull an active job_submission.
    """

    id: int
    name: str
    slurm_job_id: int
    model_config = ConfigDict(from_attributes=True, extra="ignore")


class JobSubmissionAgentSubmittedRequest(BaseModel):
    """Request model for marking JobSubmission instances as SUBMITTED."""

    id: int
    slurm_job_id: NonNegativeInt
    slurm_job_state: SlurmJobState
    slurm_job_info: str
    slurm_job_state_reason: Optional[str] = None

    model_config = ConfigDict(json_schema_extra=job_submission_meta_mapper)


class JobSubmissionAgentRejectedRequest(BaseModel):
    """Request model for marking JobSubmission instances as REJECTED."""

    id: int
    report_message: str

    model_config = ConfigDict(json_schema_extra=job_submission_meta_mapper)


class JobSubmissionAgentUpdateRequest(BaseModel):
    """Request model for updating JobSubmission instances."""

    slurm_job_id: NonNegativeInt
    slurm_job_state: SlurmJobState
    slurm_job_info: str
    slurm_job_state_reason: Optional[str] = None

    model_config = ConfigDict(json_schema_extra=job_submission_meta_mapper)
